@@ -3,6 +3,7 @@ import logging
 import warnings
 from collections import defaultdict
 from contextlib import suppress
+from ipaddress import ip_address
 from time import time
 from typing import Callable, Dict, Optional, Type, TypeVar
 
@@ -68,6 +69,9 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
         self.stats = crawler.stats
 
         self.browser_type: str = crawler.settings.get("PLAYWRIGHT_BROWSER_TYPE") or "chromium"
+        self.max_pages_per_context: int = crawler.settings.getint(
+            "PLAYWRIGHT_MAX_PAGES_PER_CONTEXT"
+        ) or crawler.settings.getint("CONCURRENT_REQUESTS")
         self.launch_options: dict = crawler.settings.getdict("PLAYWRIGHT_LAUNCH_OPTIONS") or {}
 
         self.default_navigation_timeout: Optional[float] = None
@@ -101,6 +105,8 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
             self.context_kwargs[name].update(kwargs)
         if "default" not in self.context_kwargs and default_context_kwargs:
             self.context_kwargs["default"] = default_context_kwargs
+        self.contexts: Dict[str, BrowserContext] = dict()
+        self.context_semaphores: Dict[str, asyncio.Semaphore] = dict()
 
     @classmethod
     def from_crawler(cls: Type[PlaywrightHandler], crawler: Crawler) -> PlaywrightHandler:
@@ -123,7 +129,11 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
                 for name, kwargs in self.context_kwargs.items()
             ]
         )
-        self.contexts: Dict[str, BrowserContext] = dict(zip(self.context_kwargs.keys(), contexts))
+        self.contexts = dict(zip(self.context_kwargs.keys(), contexts))
+        self.context_semaphores = {
+            name: asyncio.Semaphore(value=self.max_pages_per_context)
+            for name in self.contexts.keys()
+        }
 
     async def _create_browser_context(self, name: str, context_kwargs: dict) -> BrowserContext:
         context = await self.browser.new_context(**context_kwargs)
@@ -142,16 +152,38 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
             context_kwargs = request.meta.get("playwright_context_kwargs") or {}
             context = await self._create_browser_context(context_name, context_kwargs)
             self.contexts[context_name] = context
+            self.context_semaphores[context_name] = asyncio.Semaphore(
+                value=self.max_pages_per_context
+            )
+
+        await self.context_semaphores[context_name].acquire()
         page = await context.new_page()
+        self.stats.inc_value("playwright/page_count")
+        logger.debug(
+            "[Context=%s] New page created, page count is %i (%i for all contexts)",
+            context_name,
+            len(context.pages),
+            self._get_total_page_count(),
+        )
+        if self.default_navigation_timeout is not None:
+            page.set_default_navigation_timeout(self.default_navigation_timeout)
+
+        page.on("close", self._make_close_page_callback(context_name))
+        page.on("crash", self._make_close_page_callback(context_name))
         page.on("request", _make_request_logger(context_name))
         page.on("response", _make_response_logger(context_name))
         page.on("request", self._increment_request_stats)
         page.on("response", self._increment_response_stats)
-        self.stats.inc_value("playwright/page_count")
-        if self.default_navigation_timeout is not None:
-            page.set_default_navigation_timeout(self.default_navigation_timeout)
+
         return page
 
+    def _get_total_page_count(self):
+        count = sum([len(context.pages) for context in self.contexts.values()])
+        current_max_count = self.stats.get_value("playwright/page_count/max_concurrent")
+        if current_max_count is None or count > current_max_count:
+            self.stats.set_value("playwright/page_count/max_concurrent", count)
+        return count
+
     @inlineCallbacks
     def close(self) -> Deferred:
         yield super().close()
@@ -159,6 +191,7 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
 
     async def _close(self) -> None:
         self.contexts.clear()
+        self.context_semaphores.clear()
         if getattr(self, "browser", None):
             logger.info("Closing browser")
             await self.browser.close()
@@ -221,6 +254,10 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
                 method = getattr(page, pc.method)
                 pc.result = await method(*pc.args, **pc.kwargs)
                 await page.wait_for_load_state(timeout=self.default_navigation_timeout)
+            else:
+                logger.warning(
+                    f"Ignoring {repr(pc)}: expected PageCoroutine, got {repr(type(pc))}"
+                )
 
         body_str = await page.content()
         request.meta["download_latency"] = time() - start_time
@@ -231,6 +268,14 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
             await page.close()
             self.stats.inc_value("playwright/page_count/closed")
 
+        server_ip_address = None
+        with suppress(AttributeError, KeyError, ValueError):
+            server_addr = await response.server_addr()
+            server_ip_address = ip_address(server_addr["ipAddress"])
+
+        with suppress(AttributeError):
+            request.meta["playwright_security_details"] = await response.security_details()
+
         headers = Headers(response.headers)
         headers.pop("Content-Encoding", None)
         encoding = _get_response_encoding(headers, body_str) or "utf-8"
@@ -244,8 +289,16 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
             request=request,
             flags=["playwright"],
             encoding=encoding,
+            ip_address=server_ip_address,
         )
 
+    def _make_close_page_callback(self, context_name: str) -> Callable:
+        def close_page_callback() -> None:
+            if context_name in self.context_semaphores:
+                self.context_semaphores[context_name].release()
+
+        return close_page_callback
+
     def _increment_request_stats(self, request: PlaywrightRequest) -> None:
         stats_prefix = "playwright/request_count"
         self.stats.inc_value(stats_prefix)
@@ -265,6 +318,8 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
             logger.debug("Browser context closed: '%s'", name)
             if name in self.contexts:
                 self.contexts.pop(name)
+            if name in self.context_semaphores:
+                self.context_semaphores.pop(name)
 
         return close_browser_context_callback
 
